
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <io.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>

#include "lame.h"

#ifndef _MSC_VER
#define _set_errno(no)   (errno = (no))
#endif

/* the frame count lame decoder writes to buffer is less or equal to this */
#define LAME_BUFFER_MAX_FRAME      1152

#define FILE_BUFFER_SIZE           8192
#define WAVE_BUFFER_SIZE           (LAME_BUFFER_MAX_FRAME * 4)
#define WAVEHDR_COUNT              2

static int s_nInterrupted = 0;

static BOOL CALLBACK ConsoleIntHandler(DWORD ctrlChar)
{
	if (ctrlChar == CTRL_C_EVENT)
	{
		fprintf(stderr, "Interrupted.\n");
		s_nInterrupted = 1;
		return TRUE;
	}
	return FALSE;
}

static void tty_move_lineup(FILE* p, int lines)
{
	HANDLE h;
	CONSOLE_SCREEN_BUFFER_INFO csbi;

	h = (HANDLE) _get_osfhandle(_fileno(p));
	GetConsoleScreenBufferInfo(h, &csbi);
	csbi.dwCursorPosition.Y -= lines;
	SetConsoleCursorPosition(h, csbi.dwCursorPosition);
}

static void my_sleep(int milliseconds)
{
	Sleep((DWORD) milliseconds);
}

static int init_lame_buffers(lame_global_flags** plgf, hip_t* phip, short** pbuffL, short** pbuffR)
{
	lame_global_flags* lgf;
	hip_t hip;
	short* buffL, * buffR;

	*plgf = NULL;
	*phip = NULL;
	*pbuffL = NULL;
	*pbuffR = NULL;

	lgf = lame_init();
	if (!lgf)
	{
		perror("lametest: init_lame_buffers: lame_init");
		return -1;
	}
	/* init decoding */
	hip = hip_decode_init();
	if (!hip)
	{
		perror("lametest: init_lame_buffers: hip_decode_init");
		lame_close(lgf);
		return -1;
	}
	/* init buffer for decoding */
	buffL = (short*) malloc(sizeof(short) * LAME_BUFFER_MAX_FRAME);
	if (!buffL)
	{
		perror("lametest: init_lame_buffers: malloc[1]");
		hip_decode_exit(hip);
		lame_close(lgf);
		return -1;
	}
	buffR = (short*) malloc(sizeof(short) * LAME_BUFFER_MAX_FRAME);
	if (!buffR)
	{
		perror("lametest: init_lame_buffers: malloc[2]");
		free((void*) buffL);
		hip_decode_exit(hip);
		lame_close(lgf);
		return -1;
	}

	*plgf = lgf;
	*phip = hip;
	*pbuffL = buffL;
	*pbuffR = buffR;
	return 0;
}

static void end_lame_buffers(lame_global_flags* lgf, hip_t hip, short* buffL, short* buffR)
{
	if (buffR)
		free((void*) buffR);
	if (buffL)
		free((void*) buffL);
	if (hip)
		hip_decode_exit(hip);
	if (lgf)
		lame_close(lgf);
}

static void wave_perror(MMRESULT mr, const char* msg)
{
	char buff[MAXERRORLENGTH];

	mr = waveOutGetErrorText(mr, buff, MAXERRORLENGTH);
	if (mr != MMSYSERR_NOERROR)
		return;
	if (msg)
	{
		fprintf(stderr, msg);
		fprintf(stderr, ": ");
	}
	fprintf(stderr, buff);
	fprintf(stderr, "\n");
}

static int init_wave_api(const mp3data_struct* pmp3data, WAVEFORMATEX** ppwfx,
	WAVEHDR* pwavehdr, size_t hdrCount, HWAVEOUT* phWaveOut)
{
	MMRESULT mr;
	WAVEFORMATEX* pwfx;
	HWAVEOUT hWaveOut;
	size_t n;

	pwfx = (WAVEFORMATEX*) malloc(sizeof(WAVEFORMATEX));
	if (!pwfx)
	{
		perror("lametest: init_wave_format_ex: malloc[1]");
		*ppwfx = NULL;
		*phWaveOut = NULL;
		return -1;
	}

	/* build WAVEFORMATEX data from MP3 info */
	pwfx->wFormatTag = WAVE_FORMAT_PCM;
	pwfx->nChannels = (WORD) pmp3data->stereo;
	pwfx->nSamplesPerSec = (DWORD) pmp3data->samplerate;
	pwfx->wBitsPerSample = 16;
	pwfx->nBlockAlign = pmp3data->stereo * pwfx->wBitsPerSample / 8;
	pwfx->nAvgBytesPerSec = pwfx->nSamplesPerSec * pwfx->nBlockAlign;
	pwfx->cbSize = 0;

	/* open wave device */
	mr = waveOutOpen(&hWaveOut, WAVE_MAPPER, pwfx, 0, 0, CALLBACK_NULL);
	if (mr != MMSYSERR_NOERROR)
	{
		wave_perror(mr, "lametest: init_wave_format_ex: waveOutOpen");
		free((void*) pwfx);
		*ppwfx = NULL;
		*phWaveOut = NULL;
		return -1;
	}

	/* init WAVEHDR data */
	memset(pwavehdr, 0, sizeof(WAVEHDR) * hdrCount);
	for (n = 0; n < hdrCount; n++)
	{
		pwavehdr[n].lpData = (LPSTR) malloc(WAVE_BUFFER_SIZE);
		if (!pwavehdr[n].lpData)
		{
			perror("lametest: init_wave_format_ex: malloc[2]");
			while (n--)
			{
				waveOutUnprepareHeader(hWaveOut, &pwavehdr[n], sizeof(WAVEHDR));
				free((void*) pwavehdr[n].lpData);
				pwavehdr[n].lpData = NULL;
			}
			waveOutClose(hWaveOut);
			free(pwfx);
			*ppwfx = NULL;
			*phWaveOut = NULL;
			return -1;
		}
		mr = waveOutPrepareHeader(hWaveOut, &pwavehdr[n], sizeof(WAVEHDR));
		if (mr != MMSYSERR_NOERROR)
		{
			wave_perror(mr, "lametest: init_wave_format_ex: waveOutPrepareHeader");
			free(pwavehdr[n].lpData);
			while (n--)
			{
				waveOutUnprepareHeader(hWaveOut, &pwavehdr[n], sizeof(WAVEHDR));
				free((void*) pwavehdr[n].lpData);
				pwavehdr[n].lpData = NULL;
			}
			waveOutClose(hWaveOut);
			free(pwfx);
			*ppwfx = NULL;
			*phWaveOut = NULL;
			return -1;
		}
	}

	*ppwfx = pwfx;
	*phWaveOut = hWaveOut;
	return 0;
}

static void end_wave_api(WAVEFORMATEX* pwfx, WAVEHDR* pwavehdr, size_t hdrCount, HWAVEOUT hWaveOut)
{
	while (hdrCount--)
	{
		if (pwavehdr[hdrCount].lpData)
		{
			waveOutUnprepareHeader(hWaveOut, &pwavehdr[hdrCount], sizeof(WAVEHDR));
			free((void*) pwavehdr[hdrCount].lpData);
		}
	}
	if (hWaveOut)
		waveOutClose(hWaveOut);
	if (pwfx)
		free((void*) pwfx);
}

static size_t frame_buffer_to_wave_buffer(int frameCount, int stereo, const short* buffL, const short* buffR, void* pvBuffer, size_t _buffer_size)
{
	short* p;
	size_t n;

	if (_buffer_size < ((size_t) frameCount) * 2 * stereo)
	{
		_set_errno(EINVAL);
		return 0;
	}
	p = (short*) pvBuffer;
	n = 0;
	while (frameCount--)
	{
		*p++ = *buffL++;
		n += sizeof(short);
		if (stereo)
		{
			*p++ = *buffR++;
			n += sizeof(short);
		}
	}
	return n;
}

/* read data from file and skip ID3 data if exists */
static int read_buffer_with_skip_id3_data(FILE* pFile, void* pvBuffer, size_t* pnBufferSize, size_t* pnBufferPos)
{
	size_t nSize, nPos, headerSize;

	nSize = fread(pvBuffer, 1, FILE_BUFFER_SIZE, pFile);
	if (nSize < 4)
	{
		if (!nSize)
			perror("lametest: fread");
		else
			fprintf(stderr, "lametest: too small file size");
		return -1;
	}

	nPos = 0;
	/* check if the file has ID3 header */
	if (memcmp(pvBuffer, "ID3", 3) == 0)
	{
		if (nSize < 10)
		{
			nPos = fread(((unsigned char*) pvBuffer) + nSize, 1, 10 - nSize, pFile);
			if (nPos + nSize < 10)
			{
				if (!nPos)
					perror("lametest: fread");
				else
					fprintf(stderr, "lametest: too small file size");
				return -1;
			}

			nSize += nPos;
		}
		/* retrieve size of ID3 data */
		headerSize = ((size_t)(((unsigned char*) pvBuffer)[6] & 0x7F) << 21) |
			((size_t)(((unsigned char*) pvBuffer)[7] & 0x7F) << 14) |
			((size_t)(((unsigned char*) pvBuffer)[8] & 0x7F) << 7) |
			(size_t)(((unsigned char*) pvBuffer)[9] & 0x7F);
		/* if entire ID3 data has been read, then set current position */
		if (nSize - 10 >= headerSize)
		{
			nPos = 10 + headerSize;
			nSize -= nPos;
		}
		else
		{
			/* skip existing ID3 data */
			while (headerSize)
			{
				if (headerSize > FILE_BUFFER_SIZE)
					nSize = FILE_BUFFER_SIZE;
				else
					nSize = headerSize;
				nPos = fread(pvBuffer, 1, nSize, pFile);
				if (!nPos)
				{
					perror("lametest: fread");
					return -1;
				}
				headerSize -= nPos;
			}
			if (!headerSize)
			{
				nPos = 0;
				nSize = 0;
			}
		}
	}
	*pnBufferSize = nSize;
	*pnBufferPos = nPos;
	return 0;
}

int main(int argc, char* argv[])
{
	lame_global_flags* lgf;
	hip_t hip;
	short* buffL, * buffR;
	int nEncDelay, nEncPadding, skipSamples;
	mp3data_struct mp3data;
	void* pvBuffer;
	size_t nSize, nPos;
	FILE* pFile;

	WAVEFORMATEX* pwfx;
	HWAVEOUT hWaveOut;
	WAVEHDR wavehdr[WAVEHDR_COUNT];
	MMRESULT mr;
	int hdrPos, decodeCount, nowSample;

	int ret, exitCode;

	if (argc < 2)
	{
		fprintf(stderr, "Usage: lametest <mp3-file>\n");
		return 1;
	}

	/* set interrupted handler */
	SetConsoleCtrlHandler(ConsoleIntHandler, TRUE);

	/* init */
	ret = init_lame_buffers(&lgf, &hip, &buffL, &buffR);
	if (ret != 0)
		return 1;

	exitCode = 1;
	pvBuffer = malloc(FILE_BUFFER_SIZE);
	if (!pvBuffer)
		perror("lametest: malloc[3]");
	else
	{
		pFile = fopen(argv[1], "rb");
		if (!pFile)
			perror("lametest: fopen");
		else
		{
			ret = read_buffer_with_skip_id3_data(pFile, pvBuffer, &nSize, &nPos);
			if (ret >= 0)
			{
				decodeCount = -1;
				nEncDelay = nEncPadding = -1;
				mp3data.header_parsed = 0;
				while (1)
				{
					if (nSize)
					{
						/* retrieve the MP3 file info */
						decodeCount = hip_decode1_headersB(hip, (unsigned char*) pvBuffer + nPos, nSize,
							buffL, buffR, &mp3data, &nEncDelay, &nEncPadding);

						if (decodeCount < 0)
						{
							perror("lametest: hip_decode1_headersB");
							break;
						}
						nPos = 0;
						if (mp3data.header_parsed)
							break;
					}
					nSize = fread(pvBuffer, 1, FILE_BUFFER_SIZE, pFile);
					if (!nSize)
					{
						perror("lametest: fread");
						break;
					}
				}
				if (decodeCount >= 0)
				{
					/* calculate frames to skip */
					if (nEncDelay > -1 || nEncPadding > -1)
					{
						if (nEncDelay > -1)
							skipSamples = 528 + 1 + nEncDelay;
						if (nEncPadding > -1)
							nEncPadding -= (528 + 1);
					}
					else
						skipSamples = 528 + 1 + lame_get_encoder_delay(lgf);

					fprintf(stderr, "File %s:\n", argv[1]);
					fprintf(stderr, "  channels = %d, sample rate = %d, bit rate = %d\n"
						"  encode delay = %d, encode padding = %d, samples to skip = %d\n",
						mp3data.stereo, mp3data.samplerate, mp3data.bitrate, nEncDelay, nEncPadding, skipSamples);
					fprintf(stderr, "Press Ctrl+C to stop...\n");

					/* prepare wave out */
					ret = init_wave_api(&mp3data, &pwfx, wavehdr, WAVEHDR_COUNT, &hWaveOut);
					if (ret == 0)
					{
						hdrPos = 0;
						nowSample = 0;
						while (!s_nInterrupted)
						{
							nSize = 1;
							if (!decodeCount)
							{
								/* flush buffers that is not decoded */
								decodeCount = hip_decode1(hip, (unsigned char*) pvBuffer, 0,
									buffL, buffR);
								while (!decodeCount)
								{
									/* there is no buffer left, so read from file again */
									nSize = fread(pvBuffer, 1, FILE_BUFFER_SIZE, pFile);
									if (!nSize)
									{
										/* file is EOF, so wait until finishing the play */
										while (!(wavehdr[hdrPos].dwFlags & WHDR_DONE) && (wavehdr[hdrPos].dwFlags != WHDR_PREPARED))
										{
											if (s_nInterrupted)
												break;
											my_sleep(1);
										}
										if (!s_nInterrupted)
										{
											fprintf(stderr, "Finished.\n");
											exitCode = 0;
										}
										break;
									}
									/* decode */
									decodeCount = hip_decode1(hip, (unsigned char*) pvBuffer, nSize,
										buffL, buffR);
									if (decodeCount < 0)
									{
										perror("lametest: hip_decode1[2]");
										break;
									}
								}
							}
							if (!nSize || decodeCount < 0)
								break;
							/* skip first samples */
							if (skipSamples)
							{
								if (decodeCount < skipSamples)
								{
									skipSamples -= decodeCount;
									continue;
								}
								memmove(&buffL[0], &buffL[skipSamples], sizeof(short) * (decodeCount - skipSamples));
								memmove(&buffR[0], &buffR[skipSamples], sizeof(short) * (decodeCount - skipSamples));
								decodeCount -= skipSamples;
								skipSamples = 0;
							}
							/* skip last samples */
							if (nEncPadding > 0 && feof(pFile))
							{
								if (decodeCount < nEncPadding)
									decodeCount = 0;
								else
									decodeCount -= nEncPadding;
							}

							if (nowSample)
								tty_move_lineup(stderr, 1);
							nowSample += decodeCount;
							fprintf(stderr, "now sample = %d\n", nowSample);

							/* set frame buffer to output buffer */
							nSize = frame_buffer_to_wave_buffer(decodeCount, mp3data.stereo,
								buffL, buffR, (void*) wavehdr[hdrPos].lpData, WAVE_BUFFER_SIZE);
							if (!nSize)
							{
								perror("lametest: _frame_buffer_to_wave_buffer_s");
								break;
							}

							/* output buffer to device */
							wavehdr[hdrPos].dwBufferLength = (DWORD) nSize;
							mr = waveOutWrite(hWaveOut, &wavehdr[hdrPos], sizeof(WAVEHDR));
							if (mr != MMSYSERR_NOERROR)
							{
								wave_perror(mr, "lametest: waveOutWrite");
								break;
							}

							/* use next buffer */
							decodeCount = 0;
							hdrPos++;
							if (hdrPos == WAVEHDR_COUNT)
								hdrPos = 0;

							/* wait until the buffer can use */
							while (!(wavehdr[hdrPos].dwFlags & WHDR_DONE) && (wavehdr[hdrPos].dwFlags != WHDR_PREPARED))
							{
								if (s_nInterrupted)
									break;
								my_sleep(1);
							}
						}
						end_wave_api(pwfx, wavehdr, WAVEHDR_COUNT, hWaveOut);
					}
				}
			}
			fclose(pFile);
		}
		free(pvBuffer);
	}

	end_lame_buffers(lgf, hip, buffL, buffR);

	return exitCode;
}
